/*----------------------------------------------------------------------- 
 * 
 * File Name: ConditionData.c
 *
 * Author: Brady, P. R. and Brown, D. A.
 * 
 * Revision: $Id: ConditionData.c,v 1.43 2002/04/24 21:44:03 duncan Exp $
 * 
 *-----------------------------------------------------------------------
 */

#include "ConditionData.h"
#include <time.h>
#include <lal/Date.h>


/*
 *
 * prototypes for debugging functions
 *
 */


#ifdef INSPIRAL_SO_DEBUG
static void
PrintInspiralBank (
    LALStatus           *status,
    InspiralTemplate    *head,
    FILE                *fp
    );

static void
PrintTemplateNodeList(
    LALStatus                  *status,
    InspiralTemplateNode       *tmpltHead,
    FILE                       *fp
    );
#endif


NRCSID( CONDITIONDATAC, PKG "$Id: ConditionData.c,v 1.43 2002/04/24 21:44:03 duncan Exp $" );

/*-----------------------------------------------------------------------*/
void
LALConditionData(
    LALStatus             *status,
    LALSearchInput        *inout,
    void                  *searchParams,
    LALMPIParams          *mpiParams
    )
{
  UINT4                         i;
  INT8                          chanStartTime = 0;
  INT8                          chanStopTime  = 0;
  LIGOTimeGPS                   chanStopEpoch;
  REAL4                        *dataPtr         = NULL;
  multiDimData                 *mddTmplt        = NULL;
  multiDimData                 *mddInjEpochs    = NULL;
  multiDimData                 *mddInjParams    = NULL;
  InspiralSearchParams         *params          = NULL;
  InspiralTemplate             *currentTmplt    = NULL;
  InspiralTemplateNode         *tmpltNode       = NULL;
  REAL4TimeSeries              *chanPtr         = NULL;
  REAL4FrequencySeries         *specPtr         = NULL;
  COMPLEX8FrequencySeries      *respPtr         = NULL;


  INITSTATUS( status, "LALConditionData", CONDITIONDATAC );
  ATTATCHSTATUSPTR( status );


 /* 
  *
  * check and cast parameter and input structures
  *
  */


  ASSERT( searchParams, status,
      CONDITIONDATAH_ENULL, CONDITIONDATAH_MSGENULL );

  params = (InspiralSearchParams *) searchParams;

  chanPtr = &(params->chan);
  specPtr = &(params->spec);
  respPtr = &(params->resp);

  ASSERT (inout, status, CONDITIONDATAH_ENULL, CONDITIONDATAH_MSGENULL);


  if ( params->searchMaster )
  {


   /*
    *
    * get templates from ilwd if necessary and check that a bank exists
    *
    */


    /* templates from ilwd */
    if ( params->tmpltSource == 2 )
    {
      for ( i = 0; i < inout->numberSequences; ++i )
      {
        if ( strstr( inout->sequences[i].name, INPUTNAME_TMPLTBANK ) )
        {
          mddTmplt = &( inout->sequences[i] );
        }
      }

      if ( ! mddTmplt )
      {
        ABORT( status, CONDITIONDATAH_ENBNK, CONDITIONDATAH_MSGENBNK );
      }

      InspiralTmpltsFromILWD( status->statusPtr, params, mddTmplt );
      CHECKSTATUSPTR( status );
    }

    /* abort if we don't have a bank by now */
    if ( ! params->tmpltBankHead )
    {
      ABORT( status, CONDITIONDATAH_ENBNK, CONDITIONDATAH_MSGENBNK );
    }

#ifdef INSPIRAL_SO_DEBUG
    /* print out the template bank */
    PrintInspiralBank( status->statusPtr, params->tmpltBankHead, stdout ); 
    CHECKSTATUSPTR( status );
#endif


   /*
    *
    * create the linked list of template nodes for coarse templates
    *
    */


    /* make sure the linked list of template nodes is NULL */
    if ( params->masterParams.tmpltNodeHead )
    {
      ABORT( status, CONDITIONDATAH_ENNUL, CONDITIONDATAH_MSGENNUL );
    }

    /* we need to keep track of the number of templates for progress */
    params->masterParams.numTmpltsToFilter = 0;

    /* create the template node list of coarse templates */
    for ( currentTmplt = params->tmpltBankHead; currentTmplt; 
        currentTmplt = currentTmplt->next )
    { 
      LALFindChirpCreateTmpltNode( status->statusPtr, 
          currentTmplt, &tmpltNode );
      CHECKSTATUSPTR( status );

      /* save the head of the linked list */
      if ( ! params->masterParams.tmpltNodeHead ) 
        params->masterParams.tmpltNodeHead = tmpltNode;

      /* make sure we filter every segment against every coarse tmplt */
      for ( i = 0; i < tmpltNode->tmpltPtr->segmentIdVec->length; i++ )
        tmpltNode->tmpltPtr->segmentIdVec->data[i] = 1;

      params->masterParams.numTmpltsToFilter++;
    }

    /* record the total number of templates */
    params->masterParams.numTmpltsTotal = 
      params->masterParams.numTmpltsToFilter;

    /* if this is a bank simulation, set the number to filter to zero and */
    /* store the location of the bank head, so we can send the whole bank */
    if ( *(params->masterParams.inspiralDebugFlagPtr) == fcBankMinimalMatch )
    {
      params->masterParams.numTmpltsToFilter = 0;
      params->masterParams.tmpltBankHead = params->tmpltBankHead;
    }

    /* set the current template node to the head of the bank */
    params->masterParams.currentTmpltNode = 
      params->masterParams.tmpltNodeHead;

#ifdef INSPIRAL_SO_DEBUG
    /* print out the template node list */
    PrintTemplateNodeList( status->statusPtr, 
        params->masterParams.tmpltNodeHead, stdout );
    CHECKSTATUSPTR( status );

    fprintf( stdout, "numTmpltsToFilter = %d\n", 
        params->masterParams.numTmpltsToFilter );
    fflush( stdout );
#endif

  }
  else /* search slave */
  {


   /* 
    *
    * identify multiDimData and point the LAL structure pointers at the data
    *
    */


    for ( i = 0; i < inout->numberSequences; ++i )
    {
      if ( strstr( inout->sequences[i].name, INPUTNAME_CHANNEL ) )
      {


       /*
        *
        * data channel
        *
        */


        /* check that the length of the data channel is correct */
        if ( inout->sequences[i].range.dTime.numberSamples != 
            params->inputDataLength )
        {
          ABORT( status, CONDITIONDATAH_ECHAN, CONDITIONDATAH_MSGECHAN );
        }

        /* check the units of the data channel */
        /* XXX not yet implemented as datacondAPI does not provide units XXX */

        /* get the name of the interferometer and store in the time series */
        /* XXX this should eventually cope with channel name as well XXX   */
        if ( ! params->slaveParams.simParams )
        {
          CHAR  knownIFONames[] = KNOWN_IFO_NAMES;
          CHAR *ifoNamePtr = knownIFONames;

          memset( chanPtr->name, 0, LALNameLength * sizeof(CHAR) );

          while ( *ifoNamePtr )
          {
            CHAR ifoSearchString[] = "XX";
            strncpy( ifoSearchString, ifoNamePtr, 2 * sizeof(CHAR) );
            if ( strstr( inout->sequences[i].name, ifoSearchString ) )
            {
              strncpy ( chanPtr->name, ifoNamePtr, 2 * sizeof(CHAR) );
              break;
            }
            ifoNamePtr += 2;
          }
        }
        else
        {
          CHAR  *simIFOPtr;

          /* simulations should use S? for IFO name */
          memset( chanPtr->name, 0, LALNameLength * sizeof(CHAR) );

          if ( (simIFOPtr = strstr( inout->sequences[i].name, "S" )) )
          {
            strncpy ( chanPtr->name, simIFOPtr, 2 * sizeof(CHAR) );
          }
        }

        /* abort if we have not set an IFO name correctly */
        if ( ! chanPtr->name[0] || ! chanPtr->name[1] )
        {
          ABORT( status, CONDITIONDATAH_EUIFO, CONDITIONDATAH_MSGEUIFO );
        }

        /* set the rest of the parameters of the data channel time series */
        chanPtr->epoch.gpsSeconds = 
          inout->sequences[i].range.dTime.startSec;
        chanPtr->epoch.gpsNanoSeconds = 
          inout->sequences[i].range.dTime.startNan;
        chanPtr->deltaT = inout->sequences[i].range.dTime.timeStepSize;
        chanPtr->f0 = inout->sequences[i].range.dTime.baseFreq;
        chanPtr->sampleUnits = lalADCCountUnit;

        /* save the channel end time incase we are doing injections */
        chanStopEpoch.gpsSeconds = 
          inout->sequences[i].range.dTime.stopSec;
        chanStopEpoch.gpsNanoSeconds = 
          inout->sequences[i].range.dTime.stopNan;
        LALGPStoINT8( status->statusPtr, &chanStopTime, &chanStopEpoch );
        CHECKSTATUSPTR( status );

        /* point (or copy if simulation) the actual data channel */
        chanPtr->data->length = inout->sequences[i].range.dTime.numberSamples;
        chanPtr->data->data   = inout->sequences[i].data.real4;

        /* set sample rate for findchirp functions */
        params->slaveParams.dataParams->deltaT =
          params->slaveParams.tmpltParams->deltaT = 
          params->slaveParams.filterParams->deltaT = 
          inout->sequences[i].range.dTime.timeStepSize;

        /* store the data start time in nanoseconds */
        LALGPStoINT8( status->statusPtr, &chanStartTime, &(chanPtr->epoch) );
        CHECKSTATUSPTR( status );

        /* store a pointer to the start of the data segment */
        dataPtr = chanPtr->data->data;
      }
      else if ( strstr( inout->sequences[i].name, INPUTNAME_SPECTRUM ) )
      {


       /*
        *
        * power spectrum
        *
        */


        /* check that the length of the power spectrum is correct */
        if ( inout->sequences[i].range.dFreq.numberSamples != 
            params->initParams.numPoints / 2 + 1 )
        {
          ABORT( status, CONDITIONDATAH_ESPEC, CONDITIONDATAH_MSGESPEC );
        }

        /* check the units of the power spectrum */
        /* XXX not yet implemented as datacondAPI does not provide units XXX */

        /* set the parameters of the power spectrum frequency series */
        strncpy( specPtr->name, inout->sequences[i].name, LALNameLength );
        specPtr->epoch.gpsSeconds = 
          inout->sequences[i].range.dFreq.gpsStartTimeSec;
        specPtr->epoch.gpsNanoSeconds = 
          inout->sequences[i].range.dFreq.gpsStartTimeNan;
        specPtr->f0 = inout->sequences[i].range.dFreq.baseFreq;
        specPtr->deltaF = inout->sequences[i].range.dFreq.freqStepSize;
        {
          RAT4 negRootTwo = { -1, 1 };
          LALUnit unit;
          LALUnitPair pair;

          specPtr->sampleUnits = lalADCCountUnit;
          pair.unitOne = &lalADCCountUnit;
          pair.unitTwo = &lalHertzUnit;
          TRY( LALUnitRaise( status->statusPtr, &unit, 
                pair.unitTwo, &negRootTwo ), status );
          pair.unitTwo = &unit;
          TRY( LALUnitMultiply( status->statusPtr, &(specPtr->sampleUnits),
                &pair ), status );
        }

        /* point at the actual power spectrum */
        specPtr->data->length = inout->sequences[i].range.dFreq.numberSamples;
        specPtr->data->data   = inout->sequences[i].data.real4;
      }
      else if ( strstr( inout->sequences[i].name, INPUTNAME_RESPONSE ) )
      {


       /*
        *
        * response function
        *
        */


        /* check that the length of the response function is correct */
        if ( inout->sequences[i].range.dFreq.numberSamples != 
            params->initParams.numPoints / 2 + 1 )
        {
          ABORT( status, CONDITIONDATAH_ERESP, CONDITIONDATAH_MSGERESP );
        }

        /* check the units of the response function */
        /* XXX not yet implemented as datacondAPI does not provide units XXX */

        /* set the parameters of the response function frequency series */
        strncpy( respPtr->name, inout->sequences[i].name, LALNameLength );
        respPtr->epoch.gpsSeconds = 
          inout->sequences[i].range.dFreq.gpsStartTimeSec;
        respPtr->epoch.gpsNanoSeconds = 
          inout->sequences[i].range.dFreq.gpsStartTimeNan;
        respPtr->f0 = inout->sequences[i].range.dFreq.baseFreq;
        respPtr->deltaF = inout->sequences[i].range.dFreq.freqStepSize;
        {
          RAT4 negOne = { -1, 0 };
          LALUnit unit;
          LALUnitPair pair;

          respPtr->sampleUnits = lalStrainUnit;
          pair.unitOne = &lalStrainUnit;
          pair.unitTwo = &lalADCCountUnit;
          TRY( LALUnitRaise( status->statusPtr, &unit, 
                pair.unitTwo, &negOne ), status );
          pair.unitTwo = &unit;
          TRY( LALUnitMultiply( status->statusPtr, &(respPtr->sampleUnits),
                &pair ), status );
        }

        /* point at the actual response function data */
        respPtr->data->length = inout->sequences[i].range.dFreq.numberSamples;
        respPtr->data->data   = inout->sequences[i].data.complex8;
      }
      else if ( strstr( inout->sequences[i].name, INPUTNAME_TMPLTBANK ) )
      {
        /* we do nothing with the template bank on the slaves, but we     */
        /* have to test for it, otherwise we can't abort if there is an   */
        /* unknown channel name                                           */
      }
      else if ( strstr( inout->sequences[i].name, INPUTNAME_INJEPOCHS ) )
      {
        if ( params->slaveParams.simParams )
        {
          mddInjEpochs = &( inout->sequences[i] );
        }
      }
      else if ( strstr( inout->sequences[i].name, INPUTNAME_INJPARAMS ) )
      {
        if ( params->slaveParams.simParams )
        {
          mddInjParams = &( inout->sequences[i] );
        }
      }
      else
      {
        ABORT( status, CONDITIONDATAH_EINPT, CONDITIONDATAH_MSGEINPT ); 
      }
    }



   /* 
    *
    * create each individual DataSegment in the DataSegmentVector
    *
    */


    for ( i = 0; i < params->dataSegVec->length; ++i )
    {
      /* point to current segment */
      DataSegment      *currentSegment = params->dataSegVec->data + i;
      REAL4Sequence    *origDataPtr;

      /* this should be set to a unique number for each segment   */
      currentSegment->number = i;

      /* copy the REAL4TimeSeries information into the data segment */
      origDataPtr = currentSegment->chan->data;
      memcpy( currentSegment->chan, chanPtr, sizeof(REAL4TimeSeries) );
      currentSegment->chan->data = origDataPtr;

      /* set the correct GPS time for the current data segment */
      {
        INT8 currentSegmentTime = chanStartTime + (INT8) 
          ( 1.0e9 * 
            ((REAL8) params->initParams.numPoints - (REAL8) params->ovrlap) *
            (REAL8) i * chanPtr->deltaT
          );

        LALINT8toGPS( status->statusPtr, &(currentSegment->chan->epoch), 
            &currentSegmentTime );
        CHECKSTATUSPTR( status );
      }

      /* set up the REAL4Sequence to contain the correct data */
      currentSegment->chan->data->length = params->initParams.numPoints;
      currentSegment->chan->data->data = dataPtr;

      /* advance the dataPtr to the next segment */
      dataPtr +=  currentSegment->chan->data->length - params->ovrlap;

      /* power spectrum */
      currentSegment->spec = &(params->spec);

      /* response function */
      currentSegment->resp = &(params->resp);
    }


    /*
     *
     * perform setup for simulations
     *
     */


    if ( params->slaveParams.simParams )
    {
      if ( params->slaveParams.simParams->simType == fcBankMinimalMatch ) 
      {
#ifdef INSPIRAL_USE_URAND
        /* use the kernel random number generator to get a seed */
        FILE   *fpRand = NULL;
        INT4    randByte;

        params->seed = 0;

        fpRand = fopen( "/dev/urandom", "r" );

        if ( fpRand )
        {
          for ( randByte = 0; randByte < 4 ; ++randByte )
          {
            INT4 tmpSeed = (INT4) fgetc( fpRand );
            params->seed += tmpSeed << ( randByte * 8 );
          }
          fclose( fpRand );
        }
        else
        {
          ABORT( status, 999, "cannot open /dev/urandom" );
        }
#endif
      }

      /* create the random parameter structure */
      LALCreateRandomParams( status->statusPtr, 
          &(params->slaveParams.simParams->randomParams), params->seed );
      CHECKSTATUSPTR( status );

      if ( params->slaveParams.simParams->simType == fcRealDataInject ||
          params->slaveParams.simParams->simType == fcGaussianNoiseInject  ) 
      {
        /* parse the multiDimData containing the injection parameters */
        UINT4     numEpochs = mddInjEpochs->dimensions[1];
        UINT4     numParams = mddInjParams->dimensions[1];
        REAL4    *injParams = mddInjParams->data.real4;;

        /* XXX sanity checks missing XXX */

        /* step through the list of injection times and create an         */
        /* InspiralInjection structure for each time that lies within the */
        /* current data chunk                                             */
        for ( i = 0; i < numEpochs; ++i )
        {
          INT8                    injectTime;
          LIGOTimeGPS             injectEpoch;
          InspiralInjection      *thisEvent = NULL;

          /* get the time of the injection */
          injectEpoch.gpsSeconds     = mddInjEpochs->data.int4s[2*i];
          injectEpoch.gpsNanoSeconds = mddInjEpochs->data.int4s[2*i+1];

          LALGPStoINT8( status->statusPtr, &injectTime, &injectEpoch );
          CHECKSTATUSPTR( status );

          if ( injectTime > chanStartTime && injectTime <= chanStopTime )
          {
            if ( ! params->slaveParams.simParams->injectEvent )
            {
              thisEvent = params->slaveParams.simParams->injectEvent =
                (InspiralInjection *) LALCalloc( 1, sizeof(InspiralInjection) );
            }
            else
            {
              thisEvent = thisEvent->next = 
                (InspiralInjection *) LALCalloc( 1, sizeof(InspiralInjection) );
            }

            thisEvent->coaTime = injectTime;

            /* read the parameters for the event */
            thisEvent->totalMass    = injParams[numParams*i];
            thisEvent->eta          = injParams[numParams*i+1];
            thisEvent->dist         = injParams[numParams*i+2];
            thisEvent->inclination  = injParams[numParams*i+4];
            thisEvent->coaPhase     = injParams[numParams*i+5];
            thisEvent->longitude    = injParams[numParams*i+6];
            thisEvent->latitude     = injParams[numParams*i+7];
            thisEvent->polarization = injParams[numParams*i+8];
          }
        }
      }
    }
  }


  /* normal exit */
  DETATCHSTATUSPTR( status );
  RETURN( status );
}


#ifdef INSPIRAL_SO_DEBUG
/*-----------------------------------------------------------------------*/
static void
PrintInspiralBank (
    LALStatus           *status,
    InspiralTemplate    *head,
    FILE                *fp
    )
{
  InspiralTemplate    *current;

  INITSTATUS( status, "PrintInspiralBank", CONDITIONDATAC );
  ATTATCHSTATUSPTR( status );

  ASSERT( head, status, CONDITIONDATAH_ENULL, CONDITIONDATAH_MSGENULL );


 /*
  *
  * print the inspiral template parameter bank
  *
  */


  for ( current = head; current; current = current->next)
  {
    fprintf( fp, 
        "%p\t%d\t%f\t%f\t%f\t%f\t%f\t%p\t%p\n",
        current, current->number, 
        current->mass1, current->mass2, current->totalMass,
        current->eta, current->mu,
        current->next, current->fine );
    fflush( fp );
    if ( current->fine )
    {
      PrintInspiralBank( status->statusPtr, current->fine, fp );
      CHECKSTATUSPTR( status );
    }
  }

  DETATCHSTATUSPTR( status );
  RETURN( status );
}

/*-----------------------------------------------------------------------*/
static void
PrintTemplateNodeList(
    LALStatus                  *status,
    InspiralTemplateNode       *tmpltHead,
    FILE                       *fp
    )
{
  InspiralTemplateNode *current = NULL;

  INITSTATUS( status, "PrintTemplateNodeList", CONDITIONDATAC);

  ASSERT( tmpltHead, status, CONDITIONDATAH_ENULL, CONDITIONDATAH_MSGENULL );

  for ( current = tmpltHead; current; current = current->next )
  {
    fprintf( fp, 
        "-----------------------------------------------------------------\n" );
    fprintf( fp, "Template Node at %p points to template at %p\n",
        current, current->tmpltPtr );
    fflush( fp );
    fprintf( fp, "   number %d: m1 = %f, m2 = %f\n",
        current->tmpltPtr->number, 
        current->tmpltPtr->mass1, 
        current->tmpltPtr->mass2 );
    fflush( fp );
    fprintf( fp, "next node: %p, prev node: %p\n",
        current->next, current->prev );
    fflush( fp );
  }

  RETURN( status );
}
#endif
