/* -*- mode: c; c-basic-offset: 2; -*- */
/******** <lalVerbatim file="slopeC"> ********
Author: Daw, E
$Id: slope.c,v 1.12 2002/01/11 17:42:17 edaw Exp $
********* </lalVerbatim> ********/

#ifdef DEBUG_SLOPE
#include <stdio.h>
#endif
#include <math.h>

/* functions for memory allocation */
#include <lal/LALStdlib.h>

/* functions for vector creation and manipulation */
/* #include <lal/AVFactories.h> */

/* functions for LAL MPI interface */
#include <lal/Comm.h>

/* functions for slope filtering and peak searching data */
#include <lal/SlopeDetectorFilter.h>

/* structures and definitions for this wrapper code */
#include <Slope.h>

NRCSID( SLOPEC, "$Id: slope.c,v 1.12 2002/01/11 17:42:17 edaw Exp $");

/******** local static function declarations ********/

static
void
AddEventToLinkedList(LALStatus*        status,
		     INT4              eventTimeSecs,
                     INT4              eventTimeNSecs,
		     REAL4             eventAmp,
		     REAL4             eventDur,
		     REAL4             eventCF,
		     REAL4             eventBW,
		     SlopeDetections** pcurrentEvent,
		     SlopeDetections** pfirstEvent );

static
void
MakeEventTable(LALStatus*            status,
	       SlopeSearchParams*    params,
	       LALSearchOutput*      output);

static
void
ApplySearchMaster(LALStatus*            status,
		  LALSearchOutput*      output,
		  SlopeSearchParams*    params);

static
void
AddTimeToGpsSecNan(UINT4*     seconds,
		   UINT4*     nanoseconds,
		   REAL4      add );

/** note that currently my code is designed to run **/
/** on a single machine. ****************************/
/****************************************************/

/******** wrapperAPI function definitions ***********/

/* Initialize some parameters for the search and dynamically */
/* allocate some memory */

void LALInitSearch(LALStatus*           status,
		   void**               psearchParams,
		   LALInitSearchParams* initSearchParams)
{
  SlopeSearchParams* searchParams;
  /* these are the search input parameters */
  INT4 ac;
  CHAR** av;
  INT4 argcount;

  INITSTATUS( status, "LALInitSearch", SLOPEC );
  ATTATCHSTATUSPTR(status);

  /* get search input arguments */
  ac = initSearchParams->argc;
  av = initSearchParams->argv;

  /* allocate memory for search parameter buffer */
  searchParams = LALMalloc(sizeof(SlopeSearchParams));
  *psearchParams = (void*) searchParams;

  /* set search started bit to false */
  /* searchParams->searchStarted = 0; */

  /* initialize search parameters for simple search from defaults */

  strcpy(searchParams->channelName,
	 SLOPE_DEFAULT_CHANNEL);                    /*search channel*/

  strcpy(searchParams->ifoName,
	 SLOPE_DEFAULT_IFO);                        /*interferometer*/

  searchParams->samplingRate = 
    SLOPE_DEFAULT_SAMPLING_RATE ;                   /*sampling rate*/

  searchParams->forder = 
    SLOPE_DEFAULT_MEASURE_WIDTH ;                   /*bins for slope measure*/

  searchParams->peakWidth =                         /*number of adjacent*/
    SLOPE_DEFAULT_CORRELATED_WIDTH ;                /*filtered data bins*/
                                                    /*grouped under 1 peak*/
                                                    /*in peak search*/

  searchParams->triggerThreshold = 
    SLOPE_DEFAULT_THRESHOLD ;                       /*threshold for trigger*/

  /* go through the input paramters, overwriting some or all of */
  /* the above defaults with those present in the input parameters */

  for(argcount = 1; argcount < ac; ++argcount) {
    if (strcmp(av[argcount],"-c") == 0) {
      strcpy(searchParams->channelName,av[++argcount]);
    } else if (strcmp(av[argcount],"-smw") == 0) {
      sscanf(av[++argcount],"%u",
	     &(searchParams->forder));
    } else if (strcmp(av[argcount],"-ifo") == 0) {
      strcpy(searchParams->ifoName,av[++argcount]);
    } else if (strcmp(av[argcount],"-pw") == 0) {
      sscanf(av[++argcount],"%u",
	     &(searchParams->peakWidth));
    } else if (strcmp(av[argcount],"-sr") == 0) {
      sscanf(av[++argcount],"%f",
	     &(searchParams->samplingRate));
    } else if (strcmp(av[argcount],"-t") == 0) {
      sscanf(av[++argcount],"%f",
	     &(searchParams->triggerThreshold));
    } else if (strcmp(av[argcount],"-sdtype") == 0) {
      sscanf(av[++argcount],"%u",
	     &(searchParams->function_select));
    } else {
      ABORT(status, INITSEARCH_BADINPUTARGS, INITSEARCH_BADINPUTARGSMESG);
    }
  }
  
#ifdef SLDEBUG
  printf("%s : %u : %u : %f : %f : %u\n",
	 searchParams->channelName,
	 searchParams->forder,
	 searchParams->peakWidth,
	 searchParams->samplingRate,
	 searchParams->triggerThreshold,
	 searchParams->function_select);
#endif

  /* calculate some important quantities from the input parameters to */
  /* the search */

  searchParams->sampling_period_s = 1 / searchParams->samplingRate;
  searchParams->waveform_offset = 0;
  
  searchParams->timeIntervalForSlopeMeasure =       /*interval for slope*/
    (REAL4)searchParams->forder /                   /*measure */
    searchParams->samplingRate ;

  searchParams->numberOfTriggersSoFar = 0;          /*counter for triggers*/

  /* allocate history buffer for filter input data */
  searchParams->nSampInFiltHistBuf = 
    searchParams->forder - 1;
  searchParams->filtHistBuf = 
    LALCalloc( searchParams->nSampInFiltHistBuf,
	       sizeof(REAL4) );
  searchParams->filtHistAllocated = 0;
  searchParams->history_allocated = LALMalloc( sizeof(UINT4) );
  *(searchParams->history_allocated) = searchParams->filtHistAllocated;
  searchParams->history = searchParams->filtHistBuf;

  /* if the filter is of a convolution type, allocate memory for the taps */
  if( ((searchParams->function_select == FILTER_OUTPUT_TAPS_SET_BOXCAR) ||
       (searchParams->function_select == FILTER_OUTPUT_TAPS_SET_GAUSSIAN)) ||
      (searchParams->function_select == FILTER_OUTPUT_TAPS_SET_SINE) ) {
    searchParams->taps_set = LALMalloc( sizeof(UINT4) );
    *(searchParams->taps_set) = 0;
    searchParams->tap = LALCalloc( searchParams->forder, sizeof(REAL4));
  }

  /* allocate history buffer for peak search output data */
  searchParams->nSampInPeakHistBuf =
    searchParams->peakWidth - 1;
  searchParams->peakHistBuf =
    LALCalloc( searchParams->nSampInPeakHistBuf,
	       sizeof(REAL4) );
  searchParams->peakHistAllocated = 0;

  /* initialize parameter for length of new data set */
  searchParams->nData = 0; 

  /* initialize event list */
  searchParams->firstEvent = 0x0;
  searchParams->currentEvent = 0x0;

  /* initialize mpi information */
  searchParams->mpiMaster = initSearchParams->nodeClass;

  DETATCHSTATUSPTR(status);
  RETURN(status);
}

/* look in the data for the right channel, extract the */
/* channel data and any history buffer data to a vector */
/* of type REAL4. Get the start and end times of the data */
/* segment. */

void
LALConditionData(LALStatus*             status,
		 LALSearchInput*        inout,
		 void*                  searchParams,
		 LALMPIParams*          mpiParams)
{
  BOOLEAN dataFound = 0;
  UINT4 dCount = 0;
  UINT4 dElCount = 0;
  SlopeSearchParams* params = 0x0;
  UINT4 dataArraySize = 0;
  REAL8 historyBufferTimeInterval;
  REAL8 historyLengthSecs = 0;
  REAL8 historyLengthNSecs = 0;
#ifdef SLDEBUG
  UINT4 i;
#endif  

  INITSTATUS( status, "LALConditionData", SLOPEC );
  ATTATCHSTATUSPTR(status);
  
  params = (SlopeSearchParams*) searchParams;
  
  /*loop through the data sets in the input object and*/
  /*get the data from the channel I requested*/

  for(dCount = 0; dCount < inout->numberSequences; ++dCount) {

#ifdef SLDEBUG
    printf("%s\n",inout->sequences[dCount].name);
#endif

    if( strstr(inout->sequences[dCount].name,params->channelName) != NULL ) {
      
      /* error trap for non timeseries data here */
      if( inout->sequences[dCount].space != timeD ) {
	ABORT( status, CONDITIONDATA_NONTSDATA, CONDITIONDATA_NONTSDATAMESG);
      }
      
      dataFound = 1;
      
      /* set size of input data array including history if one exists */
      if( params->filtHistAllocated == 0 ) {
	dataArraySize = inout->sequences[dCount].dimensions[0];
      } else if( params->filtHistAllocated == 1) {
	dataArraySize = inout->sequences[dCount].dimensions[0] +
	  params->nSampInFiltHistBuf;
      } else {
	ABORT( status, CONDITIONDATA_INVALIDFILTHIST,
	       CONDITIONDATA_INVALIDFILTHISTMESG );
      }

#ifdef SLDEBUG
      printf("Size of data array is %u\n",dataArraySize);
      printf("Memory to be allocated is %u bytes.\n",dataArraySize *
	     sizeof(REAL4));
#endif

      /* allocate memory for input data */
      params->dataNew = LALCalloc( dataArraySize , sizeof(REAL4) );
      
      if( params->filtHistAllocated == 1 ) {
	/* set first elements of data vector to history buffer */
	for( dElCount = 0 ; dElCount < (params->nSampInFiltHistBuf) ;
	     ++dElCount ) {
	  params->dataNew[dElCount] = params->filtHistBuf[dElCount];
	}
      }

      /* read new data to rest of data vector. The */
      /* cast to reals depends on the original data type. For now */
      /* I support only REAL4, INT4S and INT2S */
      switch ( inout->sequences[dCount].type ) {
      case int_2s:
	if( params->filtHistAllocated == 1) {
	  for( dElCount = params->nSampInFiltHistBuf ; dElCount <
	       dataArraySize ; ++dElCount ) { 
	    params->dataNew[dElCount] = 
	      (REAL4) inout->sequences[dCount].data.int2s[dElCount 
		- params->nSampInFiltHistBuf ] ;
	  } 
	} else if (params->filtHistAllocated == 0) { 	  
	  for( dElCount = 0 ; dElCount < dataArraySize ; ++dElCount ) { 
	    params->dataNew[dElCount] = 
	      (REAL4) inout->sequences[dCount].data.int2s[dElCount] ;
	  } 
	} else {
	  ABORT( status, CONDITIONDATA_INVALIDFILTHIST,
		 CONDITIONDATA_INVALIDFILTHISTMESG );
	}
	break;
      case int_4s:
	if( params->filtHistAllocated == 1) {
	  for( dElCount = params->nSampInFiltHistBuf ; dElCount <
	       dataArraySize ; ++dElCount ) { 
	    params->dataNew[dElCount] = 
	      (REAL4) inout->sequences[dCount].data.int4s[dElCount 
		- params->nSampInFiltHistBuf ] ;
	  } 
	} else if (params->filtHistAllocated == 0) { 	  
	  for( dElCount = 0 ; dElCount < dataArraySize ; ++dElCount ) { 
	    params->dataNew[dElCount] = 
	      (REAL4) inout->sequences[dCount].data.int4s[dElCount] ;
	  } 
	} else {
	  ABORT( status, CONDITIONDATA_INVALIDFILTHIST,
		 CONDITIONDATA_INVALIDFILTHISTMESG );
	}
	break;
      case real_4:
	if( params->filtHistAllocated == 1) {
	  for( dElCount = params->nSampInFiltHistBuf ; dElCount <
	       dataArraySize ; ++dElCount ) { 
	    params->dataNew[dElCount] = 
	      inout->sequences[dCount].data.real4[dElCount 
		- params->nSampInFiltHistBuf ] ;
	  } 
	} else if (params->filtHistAllocated == 0) { 	  
	  for( dElCount = 0 ; dElCount < dataArraySize ; ++dElCount ) { 
	    params->dataNew[dElCount] = 
	      inout->sequences[dCount].data.real4[dElCount] ;
	  } 
	} else {
	  ABORT( status, CONDITIONDATA_INVALIDFILTHIST,
		 CONDITIONDATA_INVALIDFILTHISTMESG );
	}
	break;
      default:
	ABORT( status, CONDITIONDATA_DATATYPEUNSUP,
	       CONDITIONDATA_DATATYPEUNSUPMESG );
      }

      /* set data array length */
      params->nData = dataArraySize;

      if( params->nSampInFiltHistBuf > 0 ) {

	/* get duration of history buffer */
	historyBufferTimeInterval = (REAL8)params->nSampInFiltHistBuf /
	  (REAL8)params->samplingRate;

	/* split history duration into secs, nsecs. */
	historyLengthNSecs = 1e+9 * modf((REAL8)historyBufferTimeInterval,
				&historyLengthSecs);

	/* get start time of history buffer assuming it is contiguous */
	/* with the current data segment */

	if( (UINT4)historyLengthSecs > 
	    inout->sequences[dCount].range.dTime.startSec ) {
	  ABORT( status, CONDITIONDATA_TIMEARITHERROR,
		 CONDITIONDATA_TIMEARITHERRORMESG );
	} else {
	  if ( (UINT4)historyLengthNSecs <= 
	       inout->sequences[dCount].range.dTime.startNan ) {
	    /* subtract without carry */
	    params->startTimeSec = 
	      inout->sequences[dCount].range.dTime.startSec -
	      (UINT4)historyLengthSecs;
	    params->startTimeNan =
	      inout->sequences[dCount].range.dTime.startNan -
	      (UINT4)historyLengthNSecs;
	  }
	  else {
	    /* subtract with carry */
	    params->startTimeSec = 
	      inout->sequences[dCount].range.dTime.startSec -
	      (UINT4)historyLengthSecs - 1;
	    params->startTimeNan =
	      (UINT4)1000000000 + 
	      inout->sequences[dCount].range.dTime.startNan -
	      (UINT4)historyLengthNSecs;
	  }
	}
      } else {
	params->startTimeSec = 
	  inout->sequences[dCount].range.dTime.startSec -
	  (UINT4)historyLengthSecs - 1;
	params->startTimeNan =
	  (UINT4)1000000000 + 
	  inout->sequences[dCount].range.dTime.startNan -
	  (UINT4)historyLengthNSecs;
      }
    
      /* break out of for loop once data located */
      if(dataFound == 1) {
	break;
      }
      
    }
    
    if( dataFound == 0 ) {
      /* error and exit, data channel was not present */
      ABORT( status, CONDITIONDATA_NOSUCHCHANNEL,
	     CONDITIONDATA_NOSUCHCHANNELMESG );
    }

  }

#ifdef SLDEBUG
  printf("OK, reached end of condition data.\n");
#endif

  CHECKSTATUSPTR(status);
  DETATCHSTATUSPTR(status);
  RETURN(status);
}

/* apply the slope filter to the data using the slopefilters */
/* lal package. Apply a peak search algorithm to the filtered */
/* data and add an entry to the event linked list for every */
/* peak that passes the cut. Increment the output data field */
/* that keeps track of the percentage of the analysis completed. */

void
LALApplySearch(LALStatus*                status,
	       LALSearchOutput*          output,
               LALSearchInput*           input,
	       LALApplySearchParams*     applyParams)
{
  SlopeSearchParams* params;

  INITSTATUS( status, "LALApplySearch", SLOPEC );
  ATTATCHSTATUSPTR(status);

  /* extract my parameters from the search parameters */
  params = (SlopeSearchParams*)applyParams->searchParams ;

  /* figure out your rank */
  MPI_Comm_rank(*(applyParams->mpiParams->comm) , &(params->mpiRank) );

  /* if you are the master, start looking for peaks. Otherwise */
  /* abort with an error because I currently don't support multiprocessing */
  if(params->mpiMaster) {
#ifdef SLDEBUG
    printf("Applying search master routine because I am the master\n");
#endif
    ApplySearchMaster(status->statusPtr,output,params);
  } else {
    ABORT( status, APPLYSEARCH_NOSLAVES, APPLYSEARCH_NOSLAVESMESG );
  }

  /* write the linked list of located events to the */
  /* database */
#ifdef SLDEBUG
  printf("making event table.\n");
#endif

  if(params->numberOfTriggersSoFar > 0) {
    MakeEventTable(status->statusPtr, params, output);
  }

#ifdef SLDEBUG
  printf("made event table.\n");
#endif

  /* set done bit on output data */
  output->notFinished = 0;

  CHECKSTATUSPTR(status);
  DETATCHSTATUSPTR(status);
  RETURN(status);
}

/* free up memory dynamically assigned during analysis run */
/* note that I DO NOT free up memory allocated to the subfields */
/* of LALSearchOutput* output. There is no way to do this as it */
/* is not passed as an argument to LALFinalizeSearch. I am assuming */
/* that wrapperAPI tidies this up as necessary */

void
LALFinalizeSearch(LALStatus*             status,
		  void**                 psearchParams)
{
  SlopeSearchParams* searchParams = 0x0;
  SlopeDetections* pslope = 0x0;

  INITSTATUS( status, "LALFinalizeSearch", SLOPEC );
  ATTATCHSTATUSPTR(status);

  searchParams = (SlopeSearchParams*)*psearchParams;

  /* if a linked list of events was created, flush the list */

  if( searchParams->numberOfTriggersSoFar > 0 ) {
    pslope = searchParams->firstEvent;
    while( (pslope->next) != 0x0 ) {
      pslope = pslope->next;
      LALFree(pslope->previous);
    }
    LALFree( pslope );
  }

  if( ((searchParams->function_select == FILTER_OUTPUT_TAPS_SET_BOXCAR) ||
       (searchParams->function_select == FILTER_OUTPUT_TAPS_SET_GAUSSIAN)) ||
      (searchParams->function_select == FILTER_OUTPUT_TAPS_SET_SINE ) ) {
    LALFree( searchParams->taps_set );
    LALFree( searchParams->tap );
  }

  LALFree( searchParams->history_allocated );
  LALFree( searchParams->dataNew );
  LALFree( searchParams->filtHistBuf );
  LALFree( searchParams->peakHistBuf );
  LALFree( searchParams );

  DETATCHSTATUSPTR(status);
  RETURN(status);
}

/******** local static function definitions ********/

static
void
MakeEventTable(LALStatus* status,
		  SlopeSearchParams* slopeparams,
		  LALSearchOutput* output) {
  /* program name appearing in program field of SLOPE_FILTER table */

  /*this field used in Patrick Brady's moments code, not in power*/
  /*const CHAR* const PROGRAM = "slope";*/

  /* pointer to an output structure */
  outPut* poutput = 0x0;
  /* pointers to a dataBase structure */
  dataBase* db = 0x0;
  dataBase* dbprevious = 0x0;
  /* counter for events in the event list */
  UINT4 eventCounter = 0;
  /* pointer to an event list element structure */
  SlopeDetections* eventList;
  /* counter for colums in the database table */
  UINT4 dbCC=0;
  /* number of columns in database table */
  UINT4 database_column_count = 9;

  INITSTATUS( status, "MakeEventTable", SLOPEC );
  ATTATCHSTATUSPTR(status);

  /* set number of events to number of peaks found */
  output->numOutput = slopeparams->numberOfTriggersSoFar;
  /* allocate memory for database entries for each trigger */

#ifdef SLDEBUG
  printf("making output array for %u outputs \n",output->numOutput);
#endif
  poutput = LALCalloc( output->numOutput , sizeof( outPut ) ); 
  output->result = poutput;
#ifdef SLDEBUG
  printf("made output array\n");
#endif

  /* loop over events, filling the outPut structure for each event */
  eventCounter = 0;
  eventList = slopeparams->firstEvent;
  while( eventCounter < (slopeparams->numberOfTriggersSoFar)) {

#ifdef SLDEBUG
    printf("Writing event %u.\n",eventCounter);
#endif

    (poutput+eventCounter)->search           = burst;
    (poutput+eventCounter)->significant      = 1;
    (poutput+eventCounter)->states           = NULL;
    (poutput+eventCounter)->optional         = NULL;
    (poutput+eventCounter)->indexNumber      = 0;
    (poutput+eventCounter)->results = LALMalloc( sizeof(dataBase) );
    db = (poutput+eventCounter)->results;
    /* code to make a single line of the database goes here */ 
    for(dbCC = 0; dbCC<database_column_count; ++dbCC) {
      if(dbCC == 0) {
	db->previous = NULL;
      } else {
	db->previous = dbprevious;
      }
      sprintf(db->tableName,"sngl_burst");
      db->numberRows = 1;
      db->rowDimensions = NULL;
      switch (dbCC) {
      case 0:
	sprintf(db->columnName,"start_time");
	db->type = int_4s;
	db->rows.int4s = (INT4*) LALMalloc(sizeof(INT4));
	*(db->rows.int4s) = (INT4) eventList->eventDetectionTimeSeconds;
	break;
      case 1:
	sprintf(db->columnName,"start_time_ns");
	db->type = int_4s;
	db->rows.int4s = (INT4*) LALMalloc(sizeof(INT4));
	*(db->rows.int4s) = (INT4) eventList->eventDetectionTimeNanoseconds;
	break;
      case 2:
	sprintf(db->columnName,"duration");
	db->type = real_4;
	db->rows.real4 = (REAL4*) LALMalloc(sizeof(REAL4));
	*(db->rows.real4) = eventList->eventDuration;
	break;
      case 3:
	sprintf(db->columnName,"central_freq");
	db->type = real_4;
	db->rows.real4 = (REAL4*) LALMalloc(sizeof(REAL4));
	*(db->rows.real4) = eventList->eventCentralFrequency;
	break;
      case 4:
	sprintf(db->columnName,"bandwidth");
	db->type = real_4;
	db->rows.real4 = (REAL4*) LALMalloc(sizeof(REAL4));
	*(db->rows.real4) = eventList->eventBandwidth;
	break;
      case 5:
	sprintf(db->columnName,"amplitude");
	db->type = real_4;
	db->rows.real4 = (REAL4*) LALMalloc(sizeof(REAL4));
	*(db->rows.real4) = eventList->eventAmplitude;
	break;
      case 6:
	sprintf(db->columnName,"confidence");
	db->type = real_4;
	db->rows.real4 = (REAL4*) LALMalloc(sizeof(REAL4));
	*(db->rows.real4) = 0.9;
	break;
      case 7:
	sprintf(db->columnName,"search");
	db->type = char_s_ptr;
        db->rows.chars = (CHAR*) LALMalloc(6*sizeof(CHAR));
	db->rowDimensions = LALMalloc(sizeof(UINT4));
	*(db->rowDimensions) = 5;
	strcpy(db->rows.chars , "slope");
	break;
      case 8:
	sprintf(db->columnName,"ifo");
	db->type = char_s_ptr;
        db->rows.chars = (CHAR*) LALMalloc(3*sizeof(CHAR));
	db->rowDimensions = LALMalloc(sizeof(UINT4));
	*(db->rowDimensions) = 2;
	strcpy(db->rows.chars , slopeparams->ifoName);
	break;
      }
      if( dbCC == (database_column_count - 1)) {
	db->next = NULL;
      } else {
	db->next = LALMalloc( sizeof(dataBase) );
	dbprevious = db;
	db = db->next;
      }
    }
    ++eventCounter;
    eventList = (eventList->next);
#ifdef SLDEBUG
    printf("%#x\n",(unsigned int)eventList);
#endif

  }

  DETATCHSTATUSPTR(status);
  RETURN(status);
}

static
void
AddEventToLinkedList(LALStatus* const  status,
		     INT4              eventTimeSecs,
		     INT4              eventTimeNSecs,
		     REAL4             eventAmp,
		     REAL4             eventDur,
		     REAL4             eventCF,
		     REAL4             eventBW,
		     SlopeDetections** pcurrentEvent,
		     SlopeDetections** pfirstEvent ) 
{
  SlopeDetections* newEvent;
  SlopeDetections* currentEvent;

  newEvent = LALMalloc(sizeof(SlopeDetections));

  if( *pcurrentEvent == 0x0 ) {
    *pfirstEvent = newEvent;
    newEvent->previous = 0x0;
  } else {
    currentEvent = *pcurrentEvent;
    currentEvent -> next = newEvent;
    newEvent->previous = currentEvent;
    newEvent->next = 0x0;
  }

  /* allocate memory for new event. Write pointer to next field */
  /* of current event */
  /*currentEvent->next = LALMalloc( sizeof(SlopeDetections) );*/
  /*newEvent = currentEvent->next;*/

  /* fill in physical data for new event */
  newEvent->eventDetectionTimeSeconds = eventTimeSecs;
  newEvent->eventDetectionTimeNanoseconds = eventTimeNSecs;
  newEvent->eventAmplitude = eventAmp;
  newEvent->eventDuration = eventDur;
  newEvent->eventCentralFrequency = eventCF;
  newEvent->eventBandwidth = eventBW;

  /* write pointer to current event (the one passed to the function) */
  /* as the previous field of the new event */
  /*newEvent->previous = currentEvent;*/
  /*newEvent->next = 0x0;*/

  /* increment the current event to the new event */
  *pcurrentEvent = newEvent;

  RETURN(status);
}

static
void
AddTimeToGpsSecNan(UINT4*             pseconds,
		   UINT4*             pnanoseconds,
		   REAL4              add)
{
  /* this function uses math.h to break */
  /* the add time into seconds and nanoseconds. So I */
  /* need to use the double data type */
  double nsadd;
  double sadd;
  UINT4 tentonine=1000000000;
  UINT4 uint_nsadd;
  UINT4 uint_sadd;
  nsadd = 1e+09 * modf((double)add,&sadd);
  uint_nsadd = (UINT4)nsadd;
  uint_sadd = (UINT4)sadd;
  if( (*pnanoseconds) + uint_nsadd  >= tentonine ) {
    (*pseconds) = (*pseconds) + uint_sadd + 1;
    (*pnanoseconds) = ((*pnanoseconds) + uint_nsadd) % tentonine;
  } else {
    (*pseconds) = (*pseconds) + uint_sadd;
    (*pnanoseconds) = (*pnanoseconds) + uint_nsadd;
  }
  return;
}

static
void
ApplySearchMaster(LALStatus*          status,
		  LALSearchOutput*    output,
		  SlopeSearchParams*  params) 
{
  REAL4Vector dataIn;
  REAL4Vector dataSendOut;
  REAL4* dataOut;
  UINT4 dataSearchCount = 0;
  UINT4 nSearch = 0;
  UINT4 nDataOut = 0;
  UINT4 peakBin = 0;
  UINT4 peakTimeSec;
  UINT4 peakTimeNan;
  REAL4 peakAmplitude = 0.0;
  REAL4 secAfterStartStretch;
  REAL4 searchidentifier;
  SLOPEFilterParams* filterp;
  SLOPEFilterParams filterp_local;
  UINT4 hcount;
#ifdef SLDEBUG
  UINT4 i;
#endif  

#ifdef SLDEBUG
  printf("Starting apply search master.\n");
#endif

  INITSTATUS( status, "ApplySearchMaster", SLOPEC );
  ATTATCHSTATUSPTR(status);

  /* set pointer for filter parameters substructure of */
  /* the search parameters array */
  filterp = (SLOPEFilterParams*)params;

  /* get the length of the data output array */
  nDataOut = params->nData - params->forder + 1 ;
#ifdef SLDEBUG
  printf("number of bins in output array: %u\n",nDataOut);
#endif

  /* get the number of bins to search */
  nSearch = params->nData - 
    params->forder -
    params->peakWidth + 2;

  /* construct data vector for input data array. */
  dataIn.length = params->nData;
  dataIn.data = (REAL4*)params->dataNew;

  /* allocate memory for output data */
  dataOut = LALCalloc( nDataOut , sizeof(REAL4) );
  
  /* make structure containing output data array and size */
  dataSendOut.length = nDataOut;
  dataSendOut.data = dataOut;

#ifdef SLDEBUG
  printf("Starting filtration\n");
#endif

  /* set the frequency equal to the peak identifier */
  /* so I can identify which search turned up a trigger */
  /* this is a temporary kluge for E7 */
  searchidentifier = (REAL4)params->function_select;

  switch( params->function_select ) {
  case FILTER_OUTPUT_PRELIM:
    /* filter the data using the 'old' slope detector filter */
    LALSlopeDetectorFilter(status->statusPtr,
			   &dataSendOut,
			   &dataIn,
			   params->forder);
    break;
  case FILTER_OUTPUT_SLOPE:
    LALSlopeLineFitFilter(status->statusPtr,
			  &dataSendOut,
			  &dataIn,
			  *(filterp));
    break;
  case FILTER_OUTPUT_OFFSET:
    LALSlopeLineFitFilter(status->statusPtr,
			  &dataSendOut,
			  &dataIn,
			  *(filterp));
    break;
  case FILTER_OUTPUT_ALF:
    LALSlopeLineFitFilter(status->statusPtr,
			  &dataSendOut,
			  &dataIn,
			  *(filterp));
    break;
  case FILTER_OUTPUT_TAPS_SET_BOXCAR:
    if((*(filterp->taps_set)) == 0) {
      LALSlopeConvolutionFilter(status->statusPtr,
				&dataSendOut,
				&dataIn,
				*(filterp));
    }
    filterp_local.forder = filterp->forder;
    filterp_local.tap = filterp->tap;
    filterp_local.history_allocated = filterp->history_allocated;
    filterp_local.taps_set = filterp->taps_set;
    filterp_local.history = filterp->history;
    filterp_local.function_select = FILTER_OUTPUT_CONVOLVE;
    filterp_local.waveform_offset = filterp->waveform_offset;
    filterp_local.sampling_period_s = filterp->sampling_period_s;
    LALSlopeConvolutionFilter(status->statusPtr,
			      &dataSendOut,
			      &dataIn,
			      filterp_local);
    break;
  case FILTER_OUTPUT_TAPS_SET_GAUSSIAN:
    if((*(filterp->taps_set)) == 0) {
      LALSlopeConvolutionFilter(status->statusPtr,
				&dataSendOut,
				&dataIn,
				*(filterp));
    }
    filterp_local.forder = filterp->forder;
    filterp_local.tap = filterp->tap;
    filterp_local.history_allocated = filterp->history_allocated;
    filterp_local.taps_set = filterp->taps_set;
    filterp_local.history = filterp->history;
    filterp_local.function_select = FILTER_OUTPUT_CONVOLVE;
    filterp_local.waveform_offset = filterp->waveform_offset;
    filterp_local.sampling_period_s = filterp->sampling_period_s;
    LALSlopeConvolutionFilter(status->statusPtr,
			      &dataSendOut,
			      &dataIn,
			      filterp_local);
    break;
  case FILTER_OUTPUT_TAPS_SET_SINE:
    if((*(filterp->taps_set)) == 0) {
      LALSlopeConvolutionFilter(status->statusPtr,
				&dataSendOut,
				&dataIn,
				*(filterp));
    }
    filterp_local.forder = filterp->forder;
    filterp_local.tap = filterp->tap;
    filterp_local.history_allocated = filterp->history_allocated;
    filterp_local.taps_set = filterp->taps_set;
    filterp_local.history = filterp->history;
    filterp_local.function_select = FILTER_OUTPUT_CONVOLVE;
    filterp_local.waveform_offset = filterp->waveform_offset;
    filterp_local.sampling_period_s = filterp->sampling_period_s;
    LALSlopeConvolutionFilter(status->statusPtr,
			      &dataSendOut,
			      &dataIn,
			      filterp_local);
    break;
  default:
      ABORT(status, APPLYSEARCH_NOSUCHFILTER, APPLYSEARCH_NOSUCHFILTERMESG );
  }

#ifdef SLDEBUG
  printf("Starting peak search.\n");
#endif

  /* write the latter end of the output data to the output history buffer */
  for(hcount=0 ; hcount < (params->peakWidth - 1) ; ++hcount) {
    params->peakHistBuf[hcount] = dataOut[nSearch + hcount];
  }

  /* do a peak search on the filtered data */
  dataSearchCount = 0;
  while(dataSearchCount < nSearch) {

#ifdef SLDEBUG
    /* print out search information */
    printf("%u\t%f\t%f\t%f\n",dataSearchCount,
	   dataIn.data[dataSearchCount],
	   dataOut[dataSearchCount],
	   params->triggerThreshold);
#endif

    /* test if the next bin exceeds the threshold */
    if( dataOut[dataSearchCount] > params->triggerThreshold ) {

      /* if first trigger, allocate memory to first element in */
      /* linked list */
      /*if( params->numberOfTriggersSoFar == 0 ) {
	params->currentEvent = LALMalloc( sizeof(SlopeDetections) );
	params->firstEvent = params->currentEvent;
	}*/

      /* increment counter for detected events */
      ++params->numberOfTriggersSoFar;

      /* set event amplitude */
      peakAmplitude = dataOut[dataSearchCount];

      /* set bin where peak starts */
      peakBin = dataSearchCount;

      /* look over the (peakWidth) bins after the one exceeding */
      /* the threshold. The peak is positioned at the tallest one */
      while(dataSearchCount < ( peakBin + 
			     params->peakWidth - 1 )) {
	++dataSearchCount;
	if( dataOut[dataSearchCount] > peakAmplitude ) {
	  peakAmplitude = dataOut[dataSearchCount];
	  peakBin = dataSearchCount;
	}
      }

      /* get peak time */
      secAfterStartStretch = (REAL4)peakBin/((REAL4)params->samplingRate);
      peakTimeSec = params->startTimeSec;
      peakTimeNan = params->startTimeNan;
      AddTimeToGpsSecNan(&peakTimeSec, &peakTimeNan, secAfterStartStretch);

#ifdef SLDEBUG
      printf("Adding event at bin %u\n",peakBin); 
#endif

      /* write newly located peak to linked list */
      AddEventToLinkedList(status->statusPtr,
			   peakTimeSec,
			   peakTimeNan,
			   peakAmplitude,
			   ((REAL4)params->peakWidth / 
			    (REAL4)params->samplingRate),
			   searchidentifier,
			   params->samplingRate / 2,
			   &(params->currentEvent),
			   &(params->firstEvent) );

#ifdef SLDEBUG
      printf("Added event. Data search count is %u / %u\n",
	     dataSearchCount,dataSendOut.length);
    /* print out search information */
    printf("next data in %f\n",
	   dataIn.data[dataSearchCount]);
    printf("next data out %f\n",dataOut[dataSearchCount]);
#endif

    } else {
      ++dataSearchCount;
    }

    /* update progress in search */
    output->fracRemaining = ( (REAL4)(nSearch - dataSearchCount) / nSearch );
  }

#ifdef SLDEBUG
  printf("Finished search.\n");
#endif

  LALFree(dataOut);

#ifdef SLDEBUG
  printf("Freed dataOut.\n");
#endif

  DETATCHSTATUSPTR(status);
  RETURN(status);
}




