/************************** <lalVerbatim file="StochasticFinalizeSearchCV">
Author: Anderson, W.G., Romano, J.D., and Whelan, J.T.
$Id: StochasticFinalizeSearch.c,v 1.13 2002/03/30 06:02:00 whelan Exp $
**************************************************** </lalVerbatim> */

/*************************** <lalLaTeX file="StochasticFinalizeSearchC">
\subsection{Module \texttt{StochasticFinalizeSearch.c}}
\label{s:StochasticFinalizeSearch.c}

Routine that cleans up after the stochastic search.

\subsubsection*{Prototypes}
\input{StochasticFinalizeSearchCP}
\index{\texttt{LALFinalizeSearch()}}

\subsection*{Description}

Cleans up after the stochastic search--i.e., frees memory previously 
allocated in the \texttt{InitSearch} routine. 

\subsection*{Algorithms}
  
\subsection*{Uses}

\begin{itemize}
\item \texttt{LALSDestroyStochasticDataSegmentVector} 
\item \texttt{LALCDestroyStochasticDataSegmentVector} 
\item \texttt{LALCDestroyVector} 
\item \texttt{LALFree}
\item \texttt{LALDestroyRealFFTPlan}
\item \texttt{LALDestroyComplexFFTPlan}
\end{itemize}

\subsection*{Notes}

\vfill{\footnotesize\input{StochasticFinalizeSearchCV}}

*************************************************** </lalLaTeX> */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <lal/LALStdlib.h>
#include <lal/AVFactories.h>
#include <lal/TimeFreqFFT.h>
#include "StochasticData.h"
#include "StochasticFinalizeSearch.h"

NRCSID (STOCHASTICFINALIZESEARCHC, "$Id: StochasticFinalizeSearch.c,v 1.13 2002/03/30 06:02:00 whelan Exp $" );

/************************ <lalVerbatim file="StochasticFinalizeSearchCP"> */
void LALFinalizeSearch(
    LALStatus*          status,
    void**              searchParams
)
/*********************************************** </lalVerbatim> */
{
    StochasticSearchParams *params;

    
    INITSTATUS (status, "LALFinalizeSearch", STOCHASTICFINALIZESEARCHC);
    ATTATCHSTATUSPTR (status);

    
    /* 
     *
     * check and cast searchParams 
     *
     */
    
    /* :TODO: maybe an ASSERT is sufficient here?? */
    if (!(*searchParams)) {
        ABORT (status, 
               STOCHASTICFINALIZESEARCHH_ENULL, 
               STOCHASTICFINALIZESEARCHH_MSGENULL);
    }
    
    params = (StochasticSearchParams *) *searchParams;


    /*
     *
     * clean up
     *
     */

    /* free sigma, if it exists */
    if (params->sigma)
    {
      LALFree (params->sigma);
      params->sigma = NULL;
    }

    if (params->optimalFilter)
    {

      /* free optimal filter */
      TRY( LALCDestroyVector (status->statusPtr, &(params->optimalFilter->data)),
	   status );
      
      LALFree (params->optimalFilter);
      params->optimalFilter = NULL;
    
      /* free the memory used for the StochasticDataSegmentVector */
      if (!(params->initParams->baseFreq)) 
      {
	LALSDestroyStochasticDataSegmentVector 
	  (status->statusPtr, 
	   &(params->stochasticSegVec.realVec));
	CHECKSTATUSPTR (status);
	/* free fft plan */
	TRY( LALDestroyRealFFTPlan (status->statusPtr,
				    &(params->fftPlan.realPlan)),
	     status );
      }
      else 
      {
	LALCDestroyStochasticDataSegmentVector 
	  (status->statusPtr, 
	   &(params->stochasticSegVec.complexVec));
	CHECKSTATUSPTR (status);
	/* free fft plan */
	TRY( LALDestroyComplexFFTPlan (status->statusPtr,
				       &(params->fftPlan.complexPlan)),
	     status );
      }
    }
      
    /* free string fields in StochasticInitParams */

    if (params->initParams->azimuth1) {
      LALFree (params->initParams->azimuth1);
      params->initParams->azimuth1 = NULL;
    }
    if (params->initParams->azimuth2) {
      LALFree (params->initParams->azimuth2);
      params->initParams->azimuth2 = NULL;
    }

    /* free StochasticInitParams */
    LALFree (params->initParams);
    params->initParams = NULL; 

    /* free StochasticSearchParams */
    LALFree (*searchParams);
    *searchParams = NULL;

    
    /* normal exit */
    DETATCHSTATUSPTR (status);
    RETURN (status);
}
